@@ -43,6 +43,8 @@
 /* for fast user switching */
 #include <libgnomevfs/gnome-vfs-init.h>
 
+#include <libgnomekbd/gkbd-indicator.h>
+
 #include "gs-lock-plug.h"
 
 #include "gs-debug.h"
@@ -88,6 +90,8 @@ struct GSLockPlugPrivate
         GtkWidget   *switch_cancel_button;
         GtkWidget   *switch_switch_button;
 
+        GtkWidget   *auth_prompt_kbd_layout_indicator;
+
         FusaManager *fusa_manager;
 
         gboolean     caps_lock_on;
@@ -97,11 +101,9 @@ struct GSLockPlugPrivate
 
         guint        timeout;
 
-        guint        idle_id;
+        guint        cancel_timeout_id;
         guint        auth_check_idle_id;
         guint        response_idle_id;
-
-        GTimeVal     start_time;
 };
 
 typedef struct _ResponseData ResponseData;
@@ -237,20 +239,22 @@ set_status_text (GSLockPlug *plug,
         }
 }
 
-static void
-set_dialog_sensitive (GSLockPlug *plug,
-                      gboolean    sensitive)
+void
+gs_lock_plug_set_sensitive (GSLockPlug *plug,
+                            gboolean    sensitive)
 {
+        g_return_if_fail (GS_IS_LOCK_PLUG (plug));
+
         gtk_widget_set_sensitive (plug->priv->auth_prompt_entry, sensitive);
         gtk_widget_set_sensitive (plug->priv->auth_action_area, sensitive);
 }
 
 static void
-remove_monitor_idle (GSLockPlug *plug)
+remove_cancel_timeout (GSLockPlug *plug)
 {
-        if (plug->priv->idle_id > 0) {
-                g_source_remove (plug->priv->idle_id);
-                plug->priv->idle_id = 0;
+        if (plug->priv->cancel_timeout_id > 0) {
+                g_source_remove (plug->priv->cancel_timeout_id);
+                plug->priv->cancel_timeout_id = 0;
         }
 }
 
@@ -279,7 +283,7 @@ gs_lock_plug_response (GSLockPlug *plug,
                 return;
         }
 
-        remove_monitor_idle (plug);
+        remove_cancel_timeout (plug);
         remove_response_idle (plug);
 
         if (response_id == GS_LOCK_PLUG_RESPONSE_CANCEL) {
@@ -317,34 +321,21 @@ response_cancel_idle_cb (GSLockPlug *plug)
 }
 
 static gboolean
-monitor_progress (GSLockPlug *plug)
+dialog_timed_out (GSLockPlug *plug)
 {
-        GTimeVal now;
-        glong    elapsed;
-        glong    remaining;
+        gs_lock_plug_set_sensitive (plug, FALSE);
+        set_status_text (plug, _("Time has expired."));
 
-        g_get_current_time (&now);
-
-        elapsed = (now.tv_sec - plug->priv->start_time.tv_sec) * 1000
-                + (now.tv_usec - plug->priv->start_time.tv_usec) / 1000;
-        remaining = plug->priv->timeout - elapsed;
-
-        if ((remaining <= 0) || (remaining > plug->priv->timeout)) {
-                set_dialog_sensitive (plug, FALSE);
-                set_status_text (plug, _("Time has expired."));
-
-                if (plug->priv->response_idle_id != 0) {
-                        g_warning ("Response idle ID already set but shouldn't be");
-                }
-
-                remove_response_idle (plug);
-
-                plug->priv->response_idle_id = g_timeout_add (2000,
-                                                              (GSourceFunc)response_cancel_idle_cb,
-                                                              plug);
-                return FALSE;
+        if (plug->priv->response_idle_id != 0) {
+                g_warning ("Response idle ID already set but shouldn't be");
         }
-        return TRUE;
+
+        remove_response_idle (plug);
+
+        plug->priv->response_idle_id = g_timeout_add (2000,
+                                                      (GSourceFunc)response_cancel_idle_cb,
+                                                      plug);
+        return FALSE;
 }
 
 
@@ -385,14 +376,13 @@ is_capslock_on (void)
 }
 
 static void
-restart_monitor_progress (GSLockPlug *plug)
+restart_cancel_timeout (GSLockPlug *plug)
 {
-        remove_monitor_idle (plug);
+        remove_cancel_timeout (plug);
 
-        g_get_current_time (&plug->priv->start_time);
-        plug->priv->idle_id = g_timeout_add (50,
-                                             (GSourceFunc)monitor_progress,
-                                             plug);
+        plug->priv->cancel_timeout_id = g_timeout_add (plug->priv->timeout,
+                                                       (GSourceFunc)dialog_timed_out,
+                                                       plug);
 }
 
 void
@@ -496,8 +486,9 @@ gs_lock_plug_run (GSLockPlug *plug)
                 gtk_window_set_modal (GTK_WINDOW (plug), TRUE);
         }
 
-        if (!GTK_WIDGET_VISIBLE (plug))
+        if (!GTK_WIDGET_VISIBLE (plug)) {
                 gtk_widget_show (GTK_WIDGET (plug));
+        }
 
         response_handler =
                 g_signal_connect (plug,
@@ -565,7 +556,7 @@ gs_lock_plug_show (GtkWidget *widget)
 
         capslock_update (plug, is_capslock_on ());
 
-        restart_monitor_progress (plug);
+        restart_cancel_timeout (plug);
 
         gs_profile_end (NULL);
 }
@@ -659,6 +650,10 @@ gs_lock_plug_set_switch_enabled (GSLockPlug *plug,
         plug->priv->switch_enabled = switch_enabled;
         g_object_notify (G_OBJECT (plug), "switch-enabled");
 
+        if (plug->priv->auth_switch_button == NULL) {
+                return;
+        }
+
         if (switch_enabled) {
                 gtk_widget_show (plug->priv->auth_switch_button);
         } else {
@@ -813,7 +808,7 @@ switch_page (GSLockPlug *plug,
         gtk_notebook_set_current_page (GTK_NOTEBOOK (plug->priv->notebook), next_page);
 
         /* this counts as activity so restart the timer */
-        restart_monitor_progress (plug);
+        restart_cancel_timeout (plug);
 }
 
 static void
@@ -865,7 +860,11 @@ gs_lock_plug_show_prompt (GSLockPlug *plug,
         gtk_label_set_text (GTK_LABEL (plug->priv->auth_prompt_label), message);
         gtk_entry_set_visibility (GTK_ENTRY (plug->priv->auth_prompt_entry), visible);
 
-        restart_monitor_progress (plug);
+        if (! GTK_WIDGET_HAS_FOCUS (plug->priv->auth_prompt_entry)) {
+                gtk_widget_grab_focus (plug->priv->auth_prompt_entry);
+        }
+
+        restart_cancel_timeout (plug);
 }
 
 void
@@ -896,7 +895,7 @@ entry_key_press (GtkWidget   *widget,
 {
         gboolean capslock_on;
 
-        restart_monitor_progress (plug);
+        restart_cancel_timeout (plug);
 
         capslock_on = is_capslock_on ();
 
@@ -1831,6 +1830,9 @@ load_theme (GSLockPlug *plug)
         plug->priv->switch_cancel_button = glade_xml_get_widget (xml, "switch-cancel-button");
         plug->priv->switch_switch_button = glade_xml_get_widget (xml, "switch-switch-button");
 
+        /* Placeholder for the keyboard indicator */
+        plug->priv->auth_prompt_kbd_layout_indicator = glade_xml_get_widget (xml, "auth-prompt-kbd-layout-indicator");
+
         return TRUE;
 }
 
@@ -1853,7 +1855,7 @@ delete_handler (GSLockPlug  *plug,
 static void
 gs_lock_plug_init (GSLockPlug *plug)
 {
-        gunichar              invisible_char;
+        gunichar invisible_char;
 
         gs_profile_start (NULL);
 
@@ -1886,6 +1888,29 @@ gs_lock_plug_init (GSLockPlug *plug)
                 gtk_widget_show_all (plug->priv->vbox);
         }
 
+	/* Layout indicator */
+        if (plug->priv->auth_prompt_kbd_layout_indicator != NULL) {
+                XklEngine *engine;
+
+                engine = xkl_engine_get_instance (GDK_DISPLAY ());
+                if (xkl_engine_get_num_groups (engine) > 1) {
+                        GtkWidget *layout_indicator;
+
+                        layout_indicator = gkbd_indicator_new ();
+                        gkbd_indicator_set_parent_tooltips (GKBD_INDICATOR (layout_indicator), TRUE);
+                        gtk_misc_set_alignment (GTK_MISC (layout_indicator), 1, 0.5);
+                        gtk_box_pack_start (GTK_BOX (plug->priv->auth_prompt_kbd_layout_indicator),
+                                            layout_indicator, FALSE, FALSE, 6);
+
+                        gtk_widget_show_all (layout_indicator);
+                        gtk_widget_show (plug->priv->auth_prompt_kbd_layout_indicator);
+                } else {
+                        gtk_widget_hide (plug->priv->auth_prompt_kbd_layout_indicator);
+                }
+
+                g_object_unref (engine);
+	}
+
         gtk_widget_grab_default (plug->priv->auth_unlock_button);
 
         if (plug->priv->auth_username_label != NULL) {
@@ -1896,13 +1921,16 @@ gs_lock_plug_init (GSLockPlug *plug)
                 expand_string_for_label (plug->priv->auth_realname_label);
         }
 
-        if (! plug->priv->logout_enabled
-            || ! plug->priv->logout_command) {
-                gtk_widget_hide (plug->priv->auth_logout_button);
+        if (! plug->priv->logout_enabled || ! plug->priv->logout_command) {
+                if (plug->priv->auth_logout_button != NULL) {
+                        gtk_widget_hide (plug->priv->auth_logout_button);
+                }
         }
 
         if (! plug->priv->switch_enabled) {
-                gtk_widget_hide (plug->priv->auth_switch_button);
+                if (plug->priv->auth_switch_button != NULL) {
+                        gtk_widget_hide (plug->priv->auth_switch_button);
+                }
         }
 
         plug->priv->timeout = DIALOG_TIMEOUT_MSEC;
@@ -1977,7 +2005,7 @@ gs_lock_plug_finalize (GObject *object)
         }
 
         remove_response_idle (plug);
-        remove_monitor_idle (plug);
+        remove_cancel_timeout (plug);
 
         G_OBJECT_CLASS (gs_lock_plug_parent_class)->finalize (object);
 }
