@@ -47,6 +47,8 @@ enum {
         DIALOG_RESPONSE_OK
 };
 
+#define MAX_QUEUED_EVENTS 16
+
 #define GS_WINDOW_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GS_TYPE_WINDOW, GSWindowPrivate))
 
 struct GSWindowPrivate
@@ -59,11 +61,16 @@ struct GSWindowPrivate
         guint      lock_enabled : 1;
         guint      user_switch_enabled : 1;
         guint      logout_enabled : 1;
+        guint      keyboard_enabled : 1;
+
         guint64    logout_timeout;
         char      *logout_command;
+        char      *keyboard_command;
 
-        GtkWidget *box;
-        GtkWidget *socket;
+        GtkWidget *vbox;
+        GtkWidget *lock_box;
+        GtkWidget *lock_socket;
+        GtkWidget *keyboard_socket;
 
         guint      popup_dialog_idle_id;
 
@@ -73,10 +80,13 @@ struct GSWindowPrivate
 
         guint      watchdog_timer_id;
 
-        gint       pid;
-        gint       watch_id;
+        gint       lock_pid;
+        gint       lock_watch_id;
         gint       dialog_response;
 
+        gint       keyboard_pid;
+        gint       keyboard_watch_id;
+
         GList     *key_events;
 
         gdouble    last_x;
@@ -98,6 +108,8 @@ enum {
         PROP_OBSCURED,
         PROP_LOCK_ENABLED,
         PROP_LOGOUT_ENABLED,
+        PROP_KEYBOARD_ENABLED,
+        PROP_KEYBOARD_COMMAND,
         PROP_LOGOUT_COMMAND,
         PROP_LOGOUT_TIMEOUT,
         PROP_MONITOR
@@ -235,7 +247,6 @@ clear_all_children (GSWindow *window)
         clear_children (GDK_WINDOW_XID (w));
 
         gdk_display_sync (gtk_widget_get_display (GTK_WIDGET (window)));
-        gdk_flush ();
         gdk_error_trap_pop ();
 }
 
@@ -334,8 +345,84 @@ gs_window_real_unrealize (GtkWidget *widget)
         }
 }
 
+/* copied from gdk */
+extern char **environ;
+
+static gchar **
+spawn_make_environment_for_screen (GdkScreen  *screen,
+                                   gchar     **envp)
+{
+        gchar **retval = NULL;
+        gchar  *display_name;
+        gint    display_index = -1;
+        gint    i, env_len;
+
+        g_return_val_if_fail (GDK_IS_SCREEN (screen), NULL);
+
+        if (envp == NULL)
+                envp = environ;
+
+        for (env_len = 0; envp[env_len]; env_len++)
+                if (strncmp (envp[env_len], "DISPLAY", strlen ("DISPLAY")) == 0)
+                        display_index = env_len;
+
+        retval = g_new (char *, env_len + 1);
+        retval[env_len] = NULL;
+
+        display_name = gdk_screen_make_display_name (screen);
+
+        for (i = 0; i < env_len; i++)
+                if (i == display_index)
+                        retval[i] = g_strconcat ("DISPLAY=", display_name, NULL);
+                else
+                        retval[i] = g_strdup (envp[i]);
+
+        g_assert (i == env_len);
+
+        g_free (display_name);
+
+        return retval;
+}
+
+static gboolean
+spawn_command_line_on_screen_sync (GdkScreen    *screen,
+                                   const gchar  *command_line,
+                                   char        **standard_output,
+                                   char        **standard_error,
+                                   int          *exit_status,
+                                   GError      **error)
+{
+        char     **argv = NULL;
+        char     **envp = NULL;
+        gboolean   retval;
+
+        g_return_val_if_fail (command_line != NULL, FALSE);
+
+        if (! g_shell_parse_argv (command_line, NULL, &argv, error)) {
+                return FALSE;
+        }
+
+        envp = spawn_make_environment_for_screen (screen, NULL);
+
+        retval = g_spawn_sync (NULL,
+                               argv,
+                               envp,
+                               G_SPAWN_SEARCH_PATH,
+                               NULL,
+                               NULL,
+                               standard_output,
+                               standard_error,
+                               exit_status,
+                               error);
+
+        g_strfreev (argv);
+        g_strfreev (envp);
+
+        return retval;
+}
+
 static GdkVisual *
-get_best_visual (void)
+get_best_visual_for_screen (GdkScreen *screen)
 {
         char         *command;
         char         *std_output;
@@ -344,18 +431,25 @@ get_best_visual (void)
         unsigned long v;
         char          c;
         GdkVisual    *visual;
+        gboolean      res;
 
         visual = NULL;
 
         command = g_build_filename (LIBEXECDIR, "gnome-screensaver-gl-helper", NULL);
 
         error = NULL;
-        g_spawn_command_line_sync (command,
-                                   &std_output,
-                                   NULL,
-                                   &exit_status,
-                                   &error);
-        g_free (command);
+        std_output = NULL;
+        res = spawn_command_line_on_screen_sync (screen,
+                                                 command,
+                                                 &std_output,
+                                                 NULL,
+                                                 &exit_status,
+                                                 &error);
+        if (! res) {
+                gs_debug ("Could not run command '%s': %s", command, error->message);
+                g_error_free (error);
+                goto out;
+        }
 
         if (1 == sscanf (std_output, "0x%lx %c", &v, &c)) {
                 if (v != 0) {
@@ -364,12 +458,14 @@ get_best_visual (void)
                         visual_id = (VisualID) v;
                         visual = gdkx_visual_get (visual_id);
 
-                        gs_debug ("Found best visual for GL: 0x%x",
+                        gs_debug ("Found best GL visual for screen %d: 0x%x",
+                                  gdk_screen_get_number (screen),
                                   (unsigned int) visual_id);
                 }
         }
-
+ out:
         g_free (std_output);
+        g_free (command);
 
         return visual;
 }
@@ -382,7 +478,7 @@ get_best_colormap_for_screen (GdkScreen *screen)
 
         g_return_val_if_fail (screen != NULL, NULL);
 
-        visual = get_best_visual ();
+        visual = get_best_visual_for_screen (screen);
 
         colormap = NULL;
         if (visual != NULL) {
@@ -577,7 +673,8 @@ select_popup_events (void)
 
         events = SubstructureNotifyMask | attr.your_event_mask;
         XSelectInput (GDK_DISPLAY (), GDK_ROOT_WINDOW (), events);
-        gdk_flush ();
+
+        gdk_display_sync (gdk_display_get_default ());
         gdk_error_trap_pop ();
 }
 
@@ -655,6 +752,7 @@ get_env_vars (GtkWidget *widget)
         int        i;
         static const char *allowed_env_vars [] = {
                 "PATH",
+                "LD_LIBRARY_PATH",
                 "SESSION_MANAGER",
                 "XAUTHORITY",
                 "XAUTHLOCALHOSTNAME",
@@ -662,7 +760,27 @@ get_env_vars (GtkWidget *widget)
                 "KRBTKFILE",
                 "LANG",
                 "LANGUAGE",
-                "RUNNING_UNDER_GDM"
+                "LC_ALL",
+                "LC_CTYPE",
+                "LC_MESSAGES",
+                "LOGNAME",
+                "RUNNING_UNDER_GDM",
+                "GTK_DATA_PREFIX",
+                "GTK_EXE_PREFIX",
+                "GTK_IM_MODULE",
+                "GTK_IM_MODULE_FILE",
+                "GTK_MODULES",
+                "GTK_PATH",
+                "GTK2_RC_FILES",
+                "XDG_CONFIG_DIRS",
+                "XDG_DATA_DIRS",
+                "XDG_CACHE_HOME",
+                "XDG_CONFIG_HOME",
+                "XDG_DATA_HOME",
+                "MB_KBD_CONFIG",
+                "MB_KBD_VARIANT",
+                "MB_KBD_LANG",
+                "DBUS_SESSION_BUS_ADDRESS"
         };
 
         env = g_ptr_array_new ();
@@ -680,9 +798,10 @@ get_env_vars (GtkWidget *widget)
                 var = allowed_env_vars [i];
                 val = g_getenv (var);
                 if (val != NULL) {
-                        g_ptr_array_add (env, g_strdup_printf ("%s=%s",
-                                                               var,
-                                                               val));
+                        char *str;
+                        str = g_strdup_printf ("%s=%s", var, val);
+                        g_ptr_array_add (env, str);
+                        gs_debug ("adding environment: %s", str);
                 }
         }
 
@@ -691,6 +810,49 @@ get_env_vars (GtkWidget *widget)
         return env;
 }
 
+/* just for debugging */
+static gboolean
+error_watch (GIOChannel   *source,
+             GIOCondition  condition,
+             gpointer      data)
+{
+        gboolean finished = FALSE;
+
+        if (condition & G_IO_IN) {
+                GIOStatus status;
+                GError   *error = NULL;
+                char     *line;
+
+                line = NULL;
+                status = g_io_channel_read_line (source, &line, NULL, NULL, &error);
+
+                switch (status) {
+                case G_IO_STATUS_NORMAL:
+                        gs_debug ("command error output: %s", line);
+                        break;
+                case G_IO_STATUS_EOF:
+                        finished = TRUE;
+                        break;
+                case G_IO_STATUS_ERROR:
+                        finished = TRUE;
+                        gs_debug ("Error reading from child: %s\n", error->message);
+                        return FALSE;
+                case G_IO_STATUS_AGAIN:
+                default:
+                        break;
+                }
+                g_free (line);
+        } else if (condition & G_IO_HUP) {
+                finished = TRUE;
+        }
+
+        if (finished) {
+                return FALSE;
+        }
+
+        return TRUE;
+}
+
 static gboolean
 spawn_on_window (GSWindow *window,
                  char     *command,
@@ -706,13 +868,14 @@ spawn_on_window (GSWindow *window,
         gboolean    result;
         GIOChannel *channel;
         int         standard_output;
+        int         standard_error;
         int         child_pid;
         int         id;
         int         i;
 
         error = NULL;
         if (! g_shell_parse_argv (command, &argc, &argv, &error)) {
-                g_warning ("Could not parse command: %s", error->message);
+                gs_debug ("Could not parse command: %s", error->message);
                 g_error_free (error);
                 return FALSE;
         }
@@ -730,7 +893,7 @@ spawn_on_window (GSWindow *window,
                                                  &child_pid,
                                                  NULL,
                                                  &standard_output,
-                                                 NULL,
+                                                 &standard_error,
                                                  &error);
 
         for (i = 0; i < env->len; i++) {
@@ -739,18 +902,19 @@ spawn_on_window (GSWindow *window,
         g_ptr_array_free (env, TRUE);
 
         if (! result) {
-                g_warning ("Could not start command '%s': %s", command, error->message);
+                gs_debug ("Could not start command '%s': %s", command, error->message);
                 g_error_free (error);
                 g_strfreev (argv);
                 return FALSE;
         }
 
-        if (pid) {
+        if (pid != NULL) {
                 *pid = child_pid;
         } else {
                 g_spawn_close_pid (child_pid);
         }
 
+        /* output channel */
         channel = g_io_channel_unix_new (standard_output);
         g_io_channel_set_close_on_unref (channel, TRUE);
         g_io_channel_set_flags (channel,
@@ -760,10 +924,21 @@ spawn_on_window (GSWindow *window,
                              G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
                              watch_func,
                              user_data);
-        if (watch_id) {
+        if (watch_id != NULL) {
                 *watch_id = id;
         }
+        g_io_channel_unref (channel);
 
+        /* error channel */
+        channel = g_io_channel_unix_new (standard_error);
+        g_io_channel_set_close_on_unref (channel, TRUE);
+        g_io_channel_set_flags (channel,
+                                g_io_channel_get_flags (channel) | G_IO_FLAG_NONBLOCK,
+                                NULL);
+        id = g_io_add_watch (channel,
+                             G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+                             error_watch,
+                             NULL);
         g_io_channel_unref (channel);
 
         g_strfreev (argv);
@@ -772,23 +947,51 @@ spawn_on_window (GSWindow *window,
 }
 
 static void
-plug_added (GtkWidget *widget,
-            GSWindow  *window)
+lock_plug_added (GtkWidget *widget,
+                 GSWindow  *window)
 {
-        gtk_widget_show (window->priv->socket);
+        gtk_widget_show (widget);
 }
 
 static gboolean
-plug_removed (GtkWidget *widget,
-              GSWindow  *window)
+lock_plug_removed (GtkWidget *widget,
+                   GSWindow  *window)
 {
-        gtk_widget_hide (window->priv->socket);
-        gtk_container_remove (GTK_CONTAINER (window), GTK_WIDGET (window->priv->box));
-        window->priv->box = NULL;
+        gtk_widget_hide (widget);
+        gtk_container_remove (GTK_CONTAINER (window->priv->vbox), GTK_WIDGET (window->priv->lock_box));
+        window->priv->lock_box = NULL;
 
         return TRUE;
 }
 
+static void
+keyboard_plug_added (GtkWidget *widget,
+                     GSWindow  *window)
+{
+        gtk_widget_show (widget);
+}
+
+static gboolean
+keyboard_plug_removed (GtkWidget *widget,
+                       GSWindow  *window)
+{
+        gtk_widget_hide (widget);
+        gtk_container_remove (GTK_CONTAINER (window->priv->vbox), GTK_WIDGET (window->priv->keyboard_socket));
+
+        return TRUE;
+}
+
+static void
+keyboard_socket_destroyed (GtkWidget *widget,
+                           GSWindow  *window)
+{
+        g_signal_handlers_disconnect_by_func (widget, keyboard_socket_destroyed, window);
+        g_signal_handlers_disconnect_by_func (widget, keyboard_plug_added, window);
+        g_signal_handlers_disconnect_by_func (widget, keyboard_plug_removed, window);
+
+        window->priv->keyboard_socket = NULL;
+}
+
 static void
 forward_key_events (GSWindow *window)
 {
@@ -820,49 +1023,46 @@ remove_key_events (GSWindow *window)
 }
 
 static void
-socket_show (GtkWidget *widget,
-             GSWindow  *window)
+lock_socket_show (GtkWidget *widget,
+                  GSWindow  *window)
 {
-        gtk_widget_child_focus (window->priv->socket, GTK_DIR_TAB_FORWARD);
+        gtk_widget_child_focus (window->priv->lock_socket, GTK_DIR_TAB_FORWARD);
 
         /* send queued events to the dialog */
         forward_key_events (window);
 }
 
 static void
-socket_destroyed (GtkWidget *widget,
-                  GSWindow  *window)
+lock_socket_destroyed (GtkWidget *widget,
+                       GSWindow  *window)
 {
-        g_signal_handlers_disconnect_by_func (window->priv->socket, socket_show, window);
-        g_signal_handlers_disconnect_by_func (window->priv->socket, socket_destroyed, window);
-        g_signal_handlers_disconnect_by_func (window->priv->socket, plug_added, window);
-        g_signal_handlers_disconnect_by_func (window->priv->socket, plug_removed, window);
+        g_signal_handlers_disconnect_by_func (widget, lock_socket_show, window);
+        g_signal_handlers_disconnect_by_func (widget, lock_socket_destroyed, window);
+        g_signal_handlers_disconnect_by_func (widget, lock_plug_added, window);
+        g_signal_handlers_disconnect_by_func (widget, lock_plug_removed, window);
 
-        window->priv->socket = NULL;
+        window->priv->lock_socket = NULL;
 }
 
 static void
-create_socket (GSWindow *window,
-               guint32   id)
+create_keyboard_socket (GSWindow *window,
+                        guint32   id)
 {
-        window->priv->socket = gtk_socket_new ();
-        window->priv->box = gtk_alignment_new (0.5, 0.5, 0, 0);
-        gtk_widget_show (window->priv->box);
+        int height;
 
-        gtk_container_add (GTK_CONTAINER (window), window->priv->box);
+        height = (gdk_screen_get_height (gtk_widget_get_screen (GTK_WIDGET (window)))) / 4;
 
-        gtk_container_add (GTK_CONTAINER (window->priv->box), window->priv->socket);
+        window->priv->keyboard_socket = gtk_socket_new ();
+        gtk_widget_set_size_request (window->priv->keyboard_socket, -1, height);
 
-        g_signal_connect (window->priv->socket, "show",
-                          G_CALLBACK (socket_show), window);
-        g_signal_connect (window->priv->socket, "destroy",
-                          G_CALLBACK (socket_destroyed), window);
-        g_signal_connect (window->priv->socket, "plug_added",
-                          G_CALLBACK (plug_added), window);
-        g_signal_connect (window->priv->socket, "plug_removed",
-                          G_CALLBACK (plug_removed), window);
-
-        gtk_socket_add_id (GTK_SOCKET (window->priv->socket), id);
+        g_signal_connect (window->priv->keyboard_socket, "destroy",
+                          G_CALLBACK (keyboard_socket_destroyed), window);
+        g_signal_connect (window->priv->keyboard_socket, "plug_added",
+                          G_CALLBACK (keyboard_plug_added), window);
+        g_signal_connect (window->priv->keyboard_socket, "plug_removed",
+                          G_CALLBACK (keyboard_plug_removed), window);
+        gtk_box_pack_start (GTK_BOX (window->priv->vbox), window->priv->keyboard_socket, FALSE, FALSE, 0);
+        gtk_socket_add_id (GTK_SOCKET (window->priv->keyboard_socket), id);
 }
 
 /* adapted from gspawn.c */
@@ -878,33 +1078,179 @@ wait_on_child (int pid)
                 } else if (errno == ECHILD) {
                         ; /* do nothing, child already reaped */
                 } else {
-                        g_warning ("waitpid () should not fail in 'GSWindow'");
+                        gs_debug ("waitpid () should not fail in 'GSWindow'");
                 }
         }
 
         return status;
 }
 
+static void
+keyboard_command_finish (GSWindow *window)
+{
+        g_return_if_fail (GS_IS_WINDOW (window));
+
+        gs_debug ("Keyboard finished");
+
+        if (window->priv->keyboard_pid > 0) {
+                int exit_status;
+
+                exit_status = wait_on_child (window->priv->keyboard_pid);
+
+                g_spawn_close_pid (window->priv->keyboard_pid);
+                window->priv->keyboard_pid = 0;
+        }
+}
+
+static void
+kill_keyboard_command (GSWindow *window)
+{
+        if (window->priv->keyboard_pid > 0) {
+                signal_pid (window->priv->keyboard_pid, SIGTERM);
+        }
+        keyboard_command_finish (window);
+}
+
+static gboolean
+keyboard_command_watch (GIOChannel   *source,
+                        GIOCondition  condition,
+                        GSWindow     *window)
+{
+        gboolean finished = FALSE;
+
+        g_return_val_if_fail (GS_IS_WINDOW (window), FALSE);
+
+        if (condition & G_IO_IN) {
+                GIOStatus status;
+                GError   *error = NULL;
+                char     *line;
+
+                line = NULL;
+                status = g_io_channel_read_line (source, &line, NULL, NULL, &error);
+
+                switch (status) {
+                case G_IO_STATUS_NORMAL:
+                        {
+                                guint32 id;
+                                char    c;
+                                gs_debug ("keyboard command output: %s", line);
+                                if (1 == sscanf (line, " %" G_GUINT32_FORMAT " %c", &id, &c)) {
+                                        create_keyboard_socket (window, id);
+                                }
+                        }
+                        break;
+                case G_IO_STATUS_EOF:
+                        finished = TRUE;
+                        break;
+                case G_IO_STATUS_ERROR:
+                        finished = TRUE;
+                        gs_debug ("Error reading from child: %s\n", error->message);
+                        return FALSE;
+                case G_IO_STATUS_AGAIN:
+                default:
+                        break;
+                }
+
+                g_free (line);
+        } else if (condition & G_IO_HUP) {
+                finished = TRUE;
+        }
+
+        if (finished) {
+                window->priv->keyboard_watch_id = 0;
+                keyboard_command_finish (window);
+                return FALSE;
+        }
+
+        return TRUE;
+}
+
+static void
+embed_keyboard (GSWindow *window)
+{
+        gboolean res;
+
+        if (! window->priv->keyboard_enabled
+            || window->priv->keyboard_command == NULL)
+                return;
+
+        gs_debug ("Adding embedded keyboard widget");
+
+        /* FIXME: verify command is safe */
+
+        gs_debug ("Running command: %s", window->priv->keyboard_command);
+
+        res = spawn_on_window (window,
+                               window->priv->keyboard_command,
+                               &window->priv->keyboard_pid,
+                               (GIOFunc)keyboard_command_watch,
+                               window,
+                               &window->priv->keyboard_watch_id);
+        if (! res) {
+                gs_debug ("Could not start command: %s", window->priv->keyboard_command);
+        }
+}
+
+static void
+create_lock_socket (GSWindow *window,
+                    guint32   id)
+{
+        window->priv->lock_socket = gtk_socket_new ();
+        window->priv->lock_box = gtk_alignment_new (0.5, 0.5, 0, 0);
+        gtk_widget_show (window->priv->lock_box);
+        gtk_box_pack_start (GTK_BOX (window->priv->vbox), window->priv->lock_box, TRUE, TRUE, 0);
+
+        gtk_container_add (GTK_CONTAINER (window->priv->lock_box), window->priv->lock_socket);
+
+        g_signal_connect (window->priv->lock_socket, "show",
+                          G_CALLBACK (lock_socket_show), window);
+        g_signal_connect (window->priv->lock_socket, "destroy",
+                          G_CALLBACK (lock_socket_destroyed), window);
+        g_signal_connect (window->priv->lock_socket, "plug_added",
+                          G_CALLBACK (lock_plug_added), window);
+        g_signal_connect (window->priv->lock_socket, "plug_removed",
+                          G_CALLBACK (lock_plug_removed), window);
+
+        gtk_socket_add_id (GTK_SOCKET (window->priv->lock_socket), id);
+
+        if (window->priv->keyboard_enabled) {
+                embed_keyboard (window);
+        }
+}
+
 static void
 gs_window_dialog_finish (GSWindow *window)
 {
         g_return_if_fail (GS_IS_WINDOW (window));
 
         gs_debug ("Dialog finished");
+        kill_keyboard_command (window);
 
-        if (window->priv->pid > 0) {
+        if (window->priv->lock_pid > 0) {
                 int exit_status;
 
-                exit_status = wait_on_child (window->priv->pid);
+                exit_status = wait_on_child (window->priv->lock_pid);
 
-                g_spawn_close_pid (window->priv->pid);
-                window->priv->pid = 0;
+                g_spawn_close_pid (window->priv->lock_pid);
+                window->priv->lock_pid = 0;
         }
 
         /* remove events for the case were we failed to show socket */
         remove_key_events (window);
 }
 
+static void
+kill_dialog_command (GSWindow *window)
+{
+        /* If a dialog is up we need to signal it
+           and wait on it */
+        if (window->priv->lock_pid > 0) {
+                signal_pid (window->priv->lock_pid, SIGTERM);
+        }
+
+        gs_window_dialog_finish (window);
+}
+
 /* very rudimentary animation for indicating an auth failure */
 static void
 shake_dialog (GSWindow *window)
@@ -922,11 +1268,11 @@ shake_dialog (GSWindow *window)
                         right = 30;
                 }
 
-                if (! window->priv->box) {
+                if (! window->priv->lock_box) {
                         break;
                 }
 
-                gtk_alignment_set_padding (GTK_ALIGNMENT (window->priv->box),
+                gtk_alignment_set_padding (GTK_ALIGNMENT (window->priv->lock_box),
                                            0, 0,
                                            left,
                                            right);
@@ -940,9 +1286,9 @@ shake_dialog (GSWindow *window)
 }
 
 static gboolean
-command_watch (GIOChannel   *source,
-               GIOCondition  condition,
-               GSWindow     *window)
+lock_command_watch (GIOChannel   *source,
+                    GIOCondition  condition,
+                    GSWindow     *window)
 {
         gboolean finished = FALSE;
 
@@ -953,6 +1299,7 @@ command_watch (GIOChannel   *source,
                 GError   *error = NULL;
                 char     *line;
 
+                line = NULL;
                 status = g_io_channel_read_line (source, &line, NULL, NULL, &error);
 
                 switch (status) {
@@ -963,7 +1310,7 @@ command_watch (GIOChannel   *source,
                                 guint32 id;
                                 char    c;
                                 if (1 == sscanf (line, " WINDOW ID= %" G_GUINT32_FORMAT " %c", &id, &c)) {
-                                        create_socket (window, id);
+                                        create_lock_socket (window, id);
                                 }
                         } else if (strstr (line, "NOTICE=")) {
                                 if (strstr (line, "NOTICE=AUTH FAILED")) {
@@ -977,21 +1324,20 @@ command_watch (GIOChannel   *source,
                                 }
                                 finished = TRUE;
                         }
-
-                        g_free (line);
                         break;
                 case G_IO_STATUS_EOF:
                         finished = TRUE;
                         break;
                 case G_IO_STATUS_ERROR:
                         finished = TRUE;
-                        fprintf (stderr, "Error reading fd from child: %s\n", error->message);
+                        gs_debug ("Error reading from child: %s\n", error->message);
                         return FALSE;
                 case G_IO_STATUS_AGAIN:
                 default:
                         break;
                 }
 
+                g_free (line);
         } else if (condition & G_IO_HUP) {
                 finished = TRUE;
         }
@@ -1011,7 +1357,7 @@ command_watch (GIOChannel   *source,
                 window->priv->last_x = -1;
                 window->priv->last_y = -1;
 
-                window->priv->watch_id = 0;
+                window->priv->lock_watch_id = 0;
 
                 return FALSE;
         }
@@ -1078,12 +1424,12 @@ popup_dialog_idle (GSWindow *window)
 
         result = spawn_on_window (window,
                                   command->str,
-                                  &window->priv->pid,
-                                  (GIOFunc)command_watch,
+                                  &window->priv->lock_pid,
+                                  (GIOFunc)lock_command_watch,
                                   window,
-                                  &window->priv->watch_id);
+                                  &window->priv->lock_watch_id);
         if (! result) {
-                g_warning ("Could not start command: %s", command->str);
+                gs_debug ("Could not start command: %s", command->str);
         }
 
         g_string_free (command, TRUE);
@@ -1105,7 +1451,7 @@ gs_window_request_unlock (GSWindow *window)
                 return;
         }
 
-        if (window->priv->watch_id > 0) {
+        if (window->priv->lock_watch_id > 0) {
                 return;
         }
 
@@ -1155,6 +1501,30 @@ gs_window_get_screen (GSWindow  *window)
         return GTK_WINDOW (window)->screen;
 }
 
+void
+gs_window_set_keyboard_enabled (GSWindow *window,
+                                gboolean  enabled)
+{
+        g_return_if_fail (GS_IS_WINDOW (window));
+
+        window->priv->keyboard_enabled = enabled;
+}
+
+void
+gs_window_set_keyboard_command (GSWindow   *window,
+                                const char *command)
+{
+        g_return_if_fail (GS_IS_WINDOW (window));
+
+        g_free (window->priv->keyboard_command);
+
+        if (command != NULL) {
+                window->priv->keyboard_command = g_strdup (command);
+        } else {
+                window->priv->keyboard_command = NULL;
+        }
+}
+
 void
 gs_window_set_logout_enabled (GSWindow *window,
                               gboolean  logout_enabled)
@@ -1240,6 +1610,12 @@ gs_window_set_property (GObject            *object,
         case PROP_LOCK_ENABLED:
                 gs_window_set_lock_enabled (self, g_value_get_boolean (value));
                 break;
+        case PROP_KEYBOARD_ENABLED:
+                gs_window_set_keyboard_enabled (self, g_value_get_boolean (value));
+                break;
+        case PROP_KEYBOARD_COMMAND:
+                gs_window_set_keyboard_command (self, g_value_get_string (value));
+                break;
         case PROP_LOGOUT_ENABLED:
                 gs_window_set_logout_enabled (self, g_value_get_boolean (value));
                 break;
@@ -1272,6 +1648,12 @@ gs_window_get_property (GObject    *object,
         case PROP_LOCK_ENABLED:
                 g_value_set_boolean (value, self->priv->lock_enabled);
                 break;
+        case PROP_KEYBOARD_ENABLED:
+                g_value_set_boolean (value, self->priv->keyboard_enabled);
+                break;
+        case PROP_KEYBOARD_COMMAND:
+                g_value_set_string (value, self->priv->keyboard_command);
+                break;
         case PROP_LOGOUT_ENABLED:
                 g_value_set_boolean (value, self->priv->logout_enabled);
                 break;
@@ -1306,8 +1688,16 @@ queue_key_event (GSWindow    *window,
                 return;
         }
 
-        window->priv->key_events = g_list_prepend (window->priv->key_events,
-                                                   gdk_event_copy ((GdkEvent *)event));
+        /* Only cache MAX_QUEUED_EVENTS key events.  If there are any more than this then
+           something is wrong */
+        /* Don't queue keys that may cause focus navigation in the dialog */
+        if (g_list_length (window->priv->key_events) < MAX_QUEUED_EVENTS
+            && event->keyval != GDK_Tab
+            && event->keyval != GDK_Up
+            && event->keyval != GDK_Down) {
+                window->priv->key_events = g_list_prepend (window->priv->key_events,
+                                                           gdk_event_copy ((GdkEvent *)event));
+        }
 }
 
 static gboolean
@@ -1318,7 +1708,7 @@ maybe_handle_activity (GSWindow *window)
         handled = FALSE;
 
         /* if we already have a socket then don't bother */
-        if (! window->priv->socket
+        if (! window->priv->lock_socket
             && GTK_WIDGET_IS_SENSITIVE (GTK_WIDGET (window))) {
                 g_signal_emit (window, signals [ACTIVITY], 0, &handled);
         }
@@ -1330,8 +1720,6 @@ static gboolean
 gs_window_real_key_press_event (GtkWidget   *widget,
                                 GdkEventKey *event)
 {
-        gboolean catch_events = FALSE;
-
         /*g_message ("KEY PRESS state: %u keyval %u", event->state, event->keyval);*/
 
         /* Ignore brightness keys */
@@ -1340,12 +1728,9 @@ gs_window_real_key_press_event (GtkWidget   *widget,
                 return TRUE;
         }
 
-        catch_events = maybe_handle_activity (GS_WINDOW (widget));
+        maybe_handle_activity (GS_WINDOW (widget));
 
-        /* Catch all keypresses up until the lock dialog is shown */
-        if (catch_events) {
-                queue_key_event (GS_WINDOW (widget), event);
-        }
+        queue_key_event (GS_WINDOW (widget), event);
 
         if (GTK_WIDGET_CLASS (gs_window_parent_class)->key_press_event) {
                 GTK_WIDGET_CLASS (gs_window_parent_class)->key_press_event (widget, event);
@@ -1358,12 +1743,17 @@ static gboolean
 gs_window_real_motion_notify_event (GtkWidget      *widget,
                                     GdkEventMotion *event)
 {
-        GSWindow *window;
-        gdouble   distance;
-        gdouble   min_distance = 50;
+        GSWindow  *window;
+        gdouble    distance;
+        gdouble    min_distance;
+        gdouble    min_percentage = 0.1;
+        GdkScreen *screen;
 
         window = GS_WINDOW (widget);
 
+        screen = gs_window_get_screen (window);
+        min_distance = gdk_screen_get_width (screen) * min_percentage;
+
         /* if the last position was not set then don't detect motion */
         if (window->priv->last_x < 0 || window->priv->last_y < 0) {
                 window->priv->last_x = event->x;
@@ -1605,12 +1995,26 @@ gs_window_class_init (GSWindowClass *klass)
                                                             0,
                                                             G_PARAM_READWRITE));
         g_object_class_install_property (object_class,
-                                         PROP_LOGOUT_TIMEOUT,
+                                         PROP_LOGOUT_COMMAND,
                                          g_param_spec_string ("logout-command",
                                                               NULL,
                                                               NULL,
                                                               NULL,
                                                               G_PARAM_READWRITE));
+        g_object_class_install_property (object_class,
+                                         PROP_KEYBOARD_ENABLED,
+                                         g_param_spec_boolean ("keyboard-enabled",
+                                                               NULL,
+                                                               NULL,
+                                                               FALSE,
+                                                               G_PARAM_READWRITE));
+        g_object_class_install_property (object_class,
+                                         PROP_KEYBOARD_COMMAND,
+                                         g_param_spec_string ("keyboard-command",
+                                                              NULL,
+                                                              NULL,
+                                                              NULL,
+                                                              G_PARAM_READWRITE));
 
         g_object_class_install_property (object_class,
                                          PROP_MONITOR,
@@ -1658,14 +2062,22 @@ gs_window_init (GSWindow *window)
                                | GDK_ENTER_NOTIFY_MASK
                                | GDK_LEAVE_NOTIFY_MASK);
         force_no_pixmap_background (GTK_WIDGET (window));
+
+        window->priv->vbox = gtk_vbox_new (FALSE, 12);
+        gtk_widget_show (window->priv->vbox);
+        gtk_container_add (GTK_CONTAINER (window), window->priv->vbox);
 }
 
 static void
-remove_command_watch (GSWindow *window)
+remove_command_watches (GSWindow *window)
 {
-        if (window->priv->watch_id != 0) {
-                g_source_remove (window->priv->watch_id);
-                window->priv->watch_id = 0;
+        if (window->priv->lock_watch_id != 0) {
+                g_source_remove (window->priv->lock_watch_id);
+                window->priv->lock_watch_id = 0;
+        }
+        if (window->priv->keyboard_watch_id != 0) {
+                g_source_remove (window->priv->keyboard_watch_id);
+                window->priv->keyboard_watch_id = 0;
         }
 }
 
@@ -1682,6 +2094,7 @@ gs_window_finalize (GObject *object)
         g_return_if_fail (window->priv != NULL);
 
         g_free (window->priv->logout_command);
+        g_free (window->priv->keyboard_command);
 
         remove_watchdog_timer (window);
         remove_popup_dialog_idle (window);
@@ -1692,14 +2105,10 @@ gs_window_finalize (GObject *object)
 
         remove_key_events (window);
 
-        remove_command_watch (window);
+        remove_command_watches (window);
 
-        /* If a dialog is up we need to signal it
-           and wait on it */
-        if (window->priv->pid > 0) {
-                signal_pid (window->priv->pid, SIGTERM);
-        }
-        gs_window_dialog_finish (window);
+        kill_keyboard_command (window);
+        kill_dialog_command (window);
 
         G_OBJECT_CLASS (gs_window_parent_class)->finalize (object);
 }
